from pathlib import Path

import numpy as np
from skimage.io import imsave
from tifffile import imwrite

from base.process import make_rgb
from model_server.base.accessors import generate_file_accessor, InMemoryDataAccessor

class MonoPatchStack(InMemoryDataAccessor):

    def __init__(self, data):
        """
        A sequence of n monochrome images of the same size
        :param data: either np.ndarray of dimensions YXn, or a list of np.ndarrays of size YX
        """

        if isinstance(data, np.ndarray):
            if data.ndim == 3:  # interpret as YXZ
                self._data = np.expand_dims(data, 2)
            elif data.ndim == 4:  # interpret as a copy another patch stack
                self._data = data
            else:
                raise InvalidDataForPatchStackError()
        elif isinstance(data, list): # list of YX patches
            if len(data) == 0:
                self._data = np.ndarray([0, 0, 0, 0], dtype='uin9')
            elif len(data) == 1:
                self._data = np.expand_dims(
                    np.array(
                        data[0].squeeze()
                    ),
                    (2, 3)
                )
            else:
                nda = np.array(data).squeeze()
                assert nda.ndim == 3
                self._data = np.expand_dims(
                    np.moveaxis(
                        nda,
                        [1, 2, 0],
                        [0, 1, 2]),
                    2
                )
        else:
            raise InvalidDataForPatchStackError(f'Cannot create accessor from {type(data)}')

    def make_tczyx(self):
        assert self.chroma == 1
        tyx = np.moveaxis(
            self.data[:, :, 0, :], # YX(C)Z
            [2, 0, 1],
            [0, 1, 2]
        )
        return np.expand_dims(tyx, (1, 2))

    @property
    def count(self):
        return self.nz

    def iat(self, i):
        return self.data[:, :, 0, i]

    def iat_yxcz(self, i):
        return np.expand_dims(self.iat(i), (2, 3))

    def get_list(self):
        n = self.nz
        return [self.data[:, :, 0, zi] for zi in range(0, n)]


class MonoPatchStackFromFile(MonoPatchStack):
    def __init__(self, fpath):
        if not Path(fpath).exists():
            raise FileNotFoundError(f'Could not find {fpath}')
        self.file_acc = generate_file_accessor(fpath)
        super().__init__(self.file_acc.data[:, :, 0, :])

    @property
    def fpath(self):
        return self.file_acc.fpath


class PatchStack(InMemoryDataAccessor):

    def __init__(self, data):
        """
        A sequence of n (generally) color 3D images of the same size
        :param data: either a list of np.ndarrays of size YXCZ, or np.ndarray of size PYXCZ
        """

        if isinstance(data, list):  # list of YXCZ patches
            n = len(data)
            yxcz_shape = np.array([e.shape for e in data]).max(axis=0)
            nda = np.zeros(
                (n, *yxcz_shape), dtype=data[0].dtype
            )
            for i in range(0, len(data)):
                s = tuple([slice(0, c) for c in data[i].shape])
                nda[i][s] = data[i]

        elif isinstance(data, np.ndarray) and len(data.shape) == 5:  # interpret as PYXCZ
            nda = data
        else:
            raise InvalidDataForPatchStackError(f'Cannot create accessor from {type(data)}')

        assert nda.ndim == 5
        self._data = nda

    def iat(self, i):
        return self.data[i, :, :, :, :]

    def iat_yxcz(self, i):
        return self.iat(i)

    @property
    def count(self):
        return self.shape_dict['P']

    @property
    def shape_dict(self):
        return dict(zip(('P', 'Y', 'X', 'C', 'Z'), self.data.shape))

    def get_list(self):
        n = self.nz
        return [self.data[:, :, 0, zi] for zi in range(0, n)]

    @property
    def pyxcz(self):
        return self._data

    @property
    def shape(self):
        return self._data.shape

    @property
    def shape_dict(self):
        return dict(zip(('P', 'Y', 'X', 'C', 'Z'), self.data.shape))


def write_patch_to_file(where, fname, yxcz):
    ext = fname.split('.')[-1].upper()
    where.mkdir(parents=True, exist_ok=True)

    if ext == 'PNG':
        assert yxcz.dtype == 'uint8', f'Invalid data type {yxcz.dtype}'
        assert yxcz.shape[2] <= 3, f'Cannot export images with more than 3 channels as PNGs'
        assert yxcz.shape[3] == 1, f'Cannot export z-stacks as PNGs'
        if yxcz.shape[2] == 1:
            outdata = yxcz[:, :, 0, 0]
        elif yxcz.shape[2] == 2:  # add a blank blue channel
            outdata = make_rgb(yxcz)
        else: # preserve RGB order
            outdata = yxcz[:, :, :, 0]
        imsave(where / fname, outdata, check_contrast=False)
        return True

    elif ext in ['TIF', 'TIFF']:
        zcyx = np.moveaxis(yxcz, [3, 2, 0, 1], [0, 1, 2, 3])
        imwrite(where / fname, zcyx, imagej=True)
        return True

    else:
        raise Exception(f'Unsupported file extension: {ext}')


class Error(Exception):
    pass

class InvalidDataForPatchStackError(Error):
    pass

class FileNotFoundError(Error):
    pass


